! -*- mode: F90 ; mode: font-lock ; column-number-mode: true ; vc-back-end: RCS -*-
!=============================================================================!
!                          I O                                                !
!=============================================================================!
!                                                                             !
! $Id: IO.F90,v 1.11 2004/03/11 14:49:42 dq100 Exp $
!                                                                             !
!-----------------------------------------------------------------------------!
! This contains file IO stuff, most importantly the routine to convert the    !
! CASTEP.md file into a format suitable for the routines in the analysis      !
! modules.	
!                                                                             !
!-----------------------------------------------------------------------------!
! Written by D. Quigley  - University of York                                 !
!-----------------------------------------------------------------------------!
!                                                                             !
! $Log: IO.F90,v $
! Revision 1.11  2004/03/11 14:49:42  dq100
! Added XSF output option
!
! Revision 1.10  2004/02/27 13:33:11  dq100
! Tweaking of co-ordinate sytle in xmol file
!
! Revision 1.9  2004/02/27 12:25:02  dq100
! Corrected half cell shift in production.out
!
! Revision 1.8  2004/02/27 12:04:09  dq100
! Xmol now uses wrappped only, but initially translated inside cell
!
! Revision 1.7  2004/02/27 11:51:21  dq100
! Added xmol routine - wrapped co-ords only so far
!
! Revision 1.6  2004/02/19 17:37:12  dq100
! Transposed Hmatrix - correction
!
! Revision 1.5  2004/02/04 16:53:07  dq100
! Corrected temperature conversion factor
!
! Revision 1.4  2004/02/02 14:20:11  dq100
! And the errors that the correction below introduced
!
! Revision 1.3  2004/02/02 14:15:45  dq100
! Corrected output to Production.out
!
! Revision 1.2  2004/01/29 17:03:05  dq100
! Fixed lack of status on initial file opens
!
! Revision 1.1.1.1  2004/01/29 16:31:14  dq100
! Initial revision
!
!
!                                                                             !
!=============================================================================!
Module IO
  Use constants, Only : dp,int32                !Minimal useage where practical

  Implicit None                                 !Impose strong typing
  Private                                       !Everything is private ...

  !---------------------------------------------------------------------------!
  !                       P u b l i c   R o u t i n e s                       !
  !---------------------------------------------------------------------------!
  Public :: IO_ConvertFromMDFile                   ! Parses the .md file
  Public :: IO_DeleteTempFiles                     ! deletes all temp files
  Public :: IO_WriteXmolFile                       ! Writes xmol file
  Public :: IO_WriteXSFFile                        ! Writes xsf file

  ! Converts to DL_POLY FIELD and HISTORY files for use by the nMoldyn package.
  !Public :: IO_DL_PolyConv

  !---------------------------------------------------------------------------!
  !                        P u b l i c   V a r i a b l e s                    !
  !---------------------------------------------------------------------------!

  !---------------------------------------------------------------------------!
  !                      P r i v a t e   R o u t i n e s                      !
  !---------------------------------------------------------------------------!


  !---------------------------------------------------------------------------!
  !                      P r i v a t e   V a r i a b l e s                    !
  !---------------------------------------------------------------------------!

  ! Array of chemical symbols
  Character(3),Allocatable,Dimension(:) :: ElementID


Contains

  Subroutine IO_ConvertFromMDfile
    !=========================================================================!
    ! This routine calculates a radial distribution function averaged over    !
    ! all steps between the two arguments given (inclusive).                  !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !             ,  Num_Atoms   , Number of atoms in the simulation          !
    !             ,  SeedName    , Name of the simulation                     !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! The program should have been called with an input file as a command     !
    ! line argument. The first line of this should contain the seedname of    !
    ! a CASTEP run for which the file seedname.md exists.                     !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!

    Use Utility   , only : Utility_Determinant
    Use Global    , only : MD_Steps,Num_Atoms,SeedName
    Implicit None

    Real(kind=dp),Parameter :: Tconv  = 3.16681516d-6  ! a.u. to Kelvin
    Real(kind=dp),Parameter :: psconv = 0.00002417_dp   ! a.u. to ps
    Real(kind=dp),Parameter :: Pconv  = 29414000.0_dp   ! a.u. to MPa

    Real(kind=dp) :: DummyReal               ! Dummy real variable
    Real(kind=dp) :: DeltaT                  ! MD timestep
    Real(kind=dp) :: Time                    ! current time

    Real(kind=dp),Allocatable,Dimension(:) :: X,Y,Z    ! Co-ordinates
    Real(kind=dp),Allocatable,Dimension(:) :: Vx,Vy,Vz ! Velocities
    Real(kind=dp),Allocatable,Dimension(:) :: Fx,Fy,Fz ! Forces

    Real(kind=dp) :: Temperature             ! Temperature
    Real(kind=dp) :: TotalE                  ! Total Energy
    Real(kind=dp) :: HamiltE                 ! Hamiltonian Energy
    Real(kind=dp) :: KE                      ! Kinetic Energy
    Real(kind=dp) :: Pressure                ! Pressure
    Real(kind=dp) :: Volume                  ! Volume

    Real(kind=dp),Dimension(3,3) :: HMatrix  ! Matrix of Cell Vectors

    Character(3)  :: ElemID                  ! Element ID holder
    Character(80) :: mdfile                  ! name of md file
    Character(11) :: teststring              ! String used for seeking header

    Character(72) :: Dstring72               ! Dummy string
    Character(7)  :: Dstring7                !   "      "
    Character(8)  :: Dstring8                !   "      "

    Integer(kind=int32) :: R_Entries         ! Number of co-ordinate lines
    Integer(kind=int32) :: index             ! Atomic index
    Integer(kind=int32) :: ierr              ! Error flag
    Integer(kind=int32) :: I,J,Idim,Jdim     ! Loop Counters

    Logical :: Pressure_Present              ! Is pressure in MD file
    Logical :: Stress_Present                ! Is stress in MD file


    !To be killed
    character(12) :: cellfile
    character(5)  :: numstring

    ! FORTRAN formats lifted from MD.f90 by Matt Probert
1   Format(12x,es18.8e3)
2   Format(9x,3(3x,es18.8e3),'  <-- E')
3   Format(12x,es18.8,T73,   '  <-- T')
31  Format(12x,es18.8,T73,   '  <-- P')
4   Format(9x,3(3x,es18.8e3),'  <-- h')
5   Format(9x,3(3x,es18.8e3),'  <-- hv')
6   Format(9x,3(3x,es18.8e3),'  <-- S')
7   Format(1x,a3,1x,i4,3(3x,es18.8e3),'  <-- R')
8   Format(1x,a3,1x,i4,3(3x,es18.8e3),'  <-- V')
9   Format(1x,a3,1x,i4,3(3x,es18.8e3),'  <-- F')
100 Format(a72,a7)
110 Format(a72,a8)

10  Format(4F15.6)
30  Format(F15.5)
40  Format(I15)
50  Format(2I15)
60  Format(10F14.6)

    ! Open the .md file for the run to be analysed
    mdfile = Trim(seedname)//'.md'
    Open(unit=25,file=mdfile,status='old',iostat=ierr)
    If (ierr/=0) Then
       Print*,"Error - Could not open file ",Trim(mdfile)
       Stop
    End If

    ! Open a  Production.dat file
    Open(unit=26,file='Production.out',status='replace',iostat=ierr)
    If (ierr/=0) Then
       Print*,"Error - Could not open Production.out file."
       Stop
    End If

    ! Open a  Sample.dat file
    Open(unit=27,file='Sample.dat',status='replace',iostat=ierr)
    If (ierr/=0) Then
       Print*,"Error - Could not open Sample.dat file"
       Stop
    End If

    ! Open an Energy.dat file
    Open(unit=28,file='Energy.dat',status='replace',iostat=ierr)
    If (ierr/=0) Then
       Print*,"Error - Could not open Energy.dat file"
       Stop
    End If

    ! Open a  Cell.dat file
    Open(unit=29,file='Cell.dat',status='replace',iostat=ierr)
    If (ierr/=0) Then
       Print*,"Error - Could not open Cell.dat file"
       Stop
    End If

    ! Open a Forces.dat file
    Open(unit=30,file='Forces.dat',status='replace',iostat=ierr)
    If (ierr/=0) Then
       Print*,"Error - Could not open Forces.dat file"
       Stop
    End If    

    ! Read past the MD file header
    Do 
       Read(25,'(a11)',iostat=ierr) TestString
       If (ierr==0) Then
          If (TestString == ' END header' ) Then
             ! Read one more blank line
             Read(25,'(a11)')TestString
             ! Exit Loop
             Exit 
          End If
       End If
    End Do


    ! We now need to find the number of atoms and MD steps
    ! Count all <-- R entries
    R_entries = 0     ! Total no of R_entries
    Do
       Read(25,100,End=150,iostat=ierr)DString72,DString7
       If (ierr==0) Then
          If (DString7 == '  <-- R') Then
             R_entries = R_entries + 1
          End If
       End If
    End Do

150 Rewind(25)

    ! Count the number of timesteps. This should be equal to the number
    ! of occurances of the '<--E' tag.
    MD_Steps = 0
    Do
       Read(25,100,End=200,iostat=ierr)DString72,Dstring7
       If (ierr==0) Then
          If (DString7=='  <-- E') Then
             MD_Steps = MD_Steps + 1
          End If
       End If

    End Do

200 Num_Atoms = R_entries/MD_Steps


    ! Allocate the arrays
    Allocate(X(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate X array!"
       Stop
    End If
    Allocate(Y(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate Y array!"
       Stop
    End If
    Allocate(Z(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate Z array!"
       Stop
    End If
    Allocate(VX(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate VX array!"
       Stop
    End If
    Allocate(VY(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate VY array!"
       Stop
    End If
    Allocate(VZ(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate VZ array!"
       Stop
    End If
    Allocate(FX(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate FX array!"
       Stop
    End If
    Allocate(FY(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate FY array!"
       Stop
    End If
    Allocate(FZ(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate FZ array!"
       Stop
    End If


    Allocate(ElementID(1:Num_Atoms),stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_ConvertFromMDFile."
       Write(*,*)"Could not allocate ElementID array."
       Stop
    End If    

    ! Go back to start of MD file
    Rewind(25)

    ! Now we will read the first few lines to find out
    ! what information has been output.

    ! Read past the MD file header
    Do 
       Read(25,'(a11)',iostat=ierr) TestString
       If (ierr==0) Then
          If (TestString == ' END header' ) Then
             ! Read one more blank line
             Read(25,'(a11)')TestString
             ! Exit Loop
             Exit 
          End If
       End If
    End Do

    ! Should now be at start of first MD entry.
    ! so get reading - zero time
    Read(25,1)DummyReal
    Read(25,2)TotalE,HamiltE,KE
    Read(25,3)Temperature

    ! Now must deal with possibility that pressures are not output
    Read(25,100,End=200,iostat=ierr)DString72,Dstring7
    If (ierr==0) Then
       If (DString7=='  <-- P') Then
          ! This is a pressure entry
          Pressure_Present = .True.
       Else
          Backspace 25 ! Go back an entry
          Pressure_Present = .False.
       End If
    End If

    ! Next three entries are the cell vectors
    Do I = 1,3
       Read(25,4)HMatrix(I,:)
    End Do

    ! Castep has things a screwy way round
    HMatrix = transpose(HMatrix)

    ! Next entries are the cell veloctity vectors and stress
    ! which are not always present.
    Read(25,110,End=200,iostat=ierr)DString72,Dstring8
    If (ierr==0) Then
       If (DString8=='  <-- hv') Then
          ! This is a cell velocity entry
          Stress_Present = .True.
          ! Read in the rest of the entries
          Do I = 2,3
             Read(25,5)DummyReal,DummyReal,DummyReal
          End Do
          Do I = 1,3
             Read(25,6)DummyReal,DummyReal,DummyReal
          End Do
       Else
          Backspace 25 ! Go back an entry
          Stress_Present = .False.
         End If
    End If


    ! Next is the ion positions
    Do I = 1,Num_Atoms
       Read(25,7)ElemID,Index,DummyReal,DummyReal,DummyReal
    End Do

    ! Ion velocties
    Do I = 1,Num_Atoms
       Read(25,8)ElemID,Index,DummyReal,DummyReal,DummyReal
    End Do

    ! Ion forces
    Do I = 1,Num_Atoms
       Read(25,9)ElemID,Index,DummyReal,DummyReal,DummyReal
    End Do

    ! Read blank line
    Read(25,'(a11)')TestString

    ! Read the timestep
    Read(25,1)DeltaT  

    ! We can now write the header of the production.dat file
    Write(26,50)Num_Atoms,3
    Write(26,40)MD_Steps
    Write(26,40)1
    Write(26,30)DeltaT


    ! Rewind file and begin reading proper.
    Rewind(25)
    ! Read past the MD file header
    Do 
       Read(25,'(a11)',iostat=ierr) TestString
       If (ierr==0) Then
          If (TestString == ' END header' ) Then
             ! Exit Loop
             Exit 
          End If
       End If
    End Do

    ! Loop over all output timesteps
    Do J = 1,MD_Steps

       ! Read blank line
       Read(25,'(a11)')TestString

       ! Read the current time
       Read(25,1)Time

       ! Read Energies
       Read(25,2)TotalE,HamiltE,KE

       ! Read Temperature
       Read(25,3,iostat=ierr)Temperature
       if (ierr/=0) then
          Print*,"Waring - could not read temp for md step ",j
       End IF

       ! Convert to Kelvin
       Temperature = Temperature/Tconv

       If (Pressure_Present) Then
          Read(25,31)Pressure
          ! Convery to MPa
          Pressure = Pressure*Pconv
       Else
          Pressure = 0.0_dp
       End If

       ! Read cell vectors
       Do I = 1,3
          Read(25,4)HMatrix(I,:)
       End Do

       HMatrix = Transpose(HMatrix)

       Call Utility_Determinant(3,HMatrix,Volume)

       ! Write Sample.dat entry
       Write(27,10)Time*psconv,Temperature,Pressure,Volume

       ! Write Energy.dat entry
       Write(28,10)Time*psconv,KE,HamiltE,TotalE

       ! Write Cell.dat entry
       Write(29,60)Time*psconv,HMatrix(:,1),HMatrix(:,2),HMatrix(:,3)


       ! Write cell vectors to production.out file
       Do Idim = 1,3
          Write(26,*)(HMatrix(Jdim,Idim),Jdim=1,3)
       End Do

       If (Stress_Present) then
          Do I = 1,3
             Read(25,5)DummyReal,DummyReal,DummyReal
          End Do
          Do I = 1,3
             Read(25,6)DummyReal,DummyReal,DummyReal
          End Do
       End If

       ! Next is the ion positions
       Do I = 1,Num_Atoms
          Read(25,7)ElementID(I),Index,X(I),Y(I),Z(I)
       End Do

       ! Ion velocties
       Do I = 1,Num_Atoms
          Read(25,8)ElemID,Index,Vx(I),Vy(I),Vz(I)
       End Do


       ! Ion forces
       Do I = 1,Num_Atoms
          Read(25,9)ElemID,Index,Fx(I),Fy(I),Fz(I)
       End Do

       ! Write to Production.out file
       Do I = 1,Num_Atoms
          Write(26,*)X(I),Y(I),Z(I)
          Write(26,*)Vx(I),Vy(I),Vz(I)         
       End Do

       ! Write to Forces.dat file
       Do I = 1,Num_Atoms
          Write(30,*)FX(I),FY(I),FZ(I)
       End Do

    End Do

    ! Have now created a Sample.dat, Production.out, Energy.dat and
    ! Sample.dat. The user may wish to keep these for plotting
    ! purposes.....
    Close(25)
    Close(26)
    Close(27)
    Close(28)
    Close(29)
    Close(30)

    Return

  End Subroutine IO_ConvertFromMDFile

  Subroutine IO_DeleteTempFiles
    !=========================================================================!
    ! Deletes the temporary files                                             ! 
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!
    Implicit None

    Integer(kind=int32) :: ierr

    ! Check that the files exist so they can be deleted

    Open(23,File='Sample.dat',status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_DeleteTempFiles!"  
       Write(*,*)"Could not open Sample.dat file."
       Stop
    Else
       Close(23,status='delete')
    End If
    Open(23,File='Energy.dat',status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_DeleteTempFiles!"  
       Write(*,*)"Could not open Energy.dat file."
       Stop
    Else
       Close(23,status='delete')
    End If
    Open(23,File='Cell.dat',status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_DeleteTempFiles!"  
       Write(*,*)"Could not open Cell.dat file."
       Stop
    Else
       Close(23,status='delete')
    End If
    Open(23,File='Production.out',status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_DeleteTempFiles!"  
       Write(*,*)"Could not open Production.out file."
       Stop
    Else
       Close(23,status='delete')
    End If
    Open(23,File='Forces.dat',status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in IO_DeleteTempFiles!"  
       Write(*,*)"Could not open Forces.dat file."
       Stop
    Else
       Close(23,status='delete')
    End If

    return

  End Subroutine IO_DeleteTempFiles


  Subroutine IO_WriteXSFFile
    !=========================================================================!
    ! Creates an XSF file suitable for use with the Xcrystden package         !                    !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !   ElementID, String with chemical symbol for element used.              !
    !   SeedName , ID string for this run taken from input filename.          !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    !   Global,                                                               !
    !   MD_Steps          , Number of output steps                            !
    !   Num_Atoms         , Number of atoms                                   !
    !   SeedName          , Calculation Seedname                              !
    !                                                                         !
    !   Utility,                                                              !
    !   CalcRFrac         , Calculates fractional co-ordinates                !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    !-------------------------------------------------------------------------!
    ! Written by David Quigley  Feb 2004                                      !
    !=========================================================================!
    Use Global    , only : Num_Atoms,MD_Steps,SeedName
    Use Utility   , only : Utility_CalcRFrac
    Implicit None

    ! Loop counters
    Integer(kind=int32) :: I,J,Idim,Jdim

    Integer(kind=int32) :: Itmp       ! Temporary integer place holder
    Integer(kind=int32) :: ierr       ! I/O error flag

    !Holders for co-ordinates
    Real(kind=dp),Allocatable,Dimension(:,:) :: R,F
    Real(kind=dp),Allocatable,Dimension(:,:) :: RFrac 

    Real(kind=dp),Allocatable,Dimension(:,:) :: R0unwrap,R0wrap,Rdiff

    Real(kind=dp),Dimension(1:3,1:3) :: HMatrix

    ! Filename for Xmol file
    Character(30) :: XSFFile

    ! Conversion from internal units to Angstroms
    Real(kind=dp),Parameter :: BohrToAng = 0.5292_dp

    Real(kind=dp) :: tmp


100 Format(A3,6F12.6)
200 Format(3F12.6)

    Allocate(R(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*) "Error in IO_WriteXSFFile - Could not allocate R"
       Stop
    end if
    Allocate(F(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*) "Error in IO_WriteXSFFile - Could not allocate F"
       Stop
    end if
    Allocate(RFrac(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*)"Error in IO_WriteXSFFile - Could not allocate RFrac"
       Stop
    end if
    Allocate(R0unwrap(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*)"Error in IO_WriteXSFFile - Could not allocate R0unwrap"
       Stop
    end if
    Allocate(R0wrap(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*)"Error in IO_WriteXSFFile - Could not allocate R0wrap"
       Stop
    end if
    Allocate(Rdiff(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*)"Error in IO_WriteXSFFile - Could not allocate Rdiff"
       Stop
    end if
    Open(unit=29,file="Production.out",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error opening Production.out file for reading."
       Stop
    End if
    Open(unit=30,file="Forces.dat",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error opening Forces.dat file for reading."
       Stop
    End if



    ! Specify output filename
    XSFFile  = trim(SeedName)//'.axsf'

    Open(31,file=trim(XSFFile),status='replace',iostat=ierr)
    If (ierr.NE.0) then
       Write(*,*)'Error opening '//xsffile//'for output'
       Stop
    End If


    ! Skip past header
    Read(29,'(2i15)' )Num_Atoms,Itmp  ! Number of atoms,Dimensionality
    Read(29,'(i15)'  )MD_Steps        ! Number of outputs in file
    Read(29,'(i15)'  )Itmp            ! Interval in timesteps
    Read(29,'(f15.5)')tmp             ! Calculation Timestep

    ! Write the xfs header
    Write(31,'(a10,i5)')'ANIMSTEPS ',MD_Steps
    
    ! Always allow 3d periodicity for now
    Write(31,*)'CRYSTAL'

    ! Create the Xmol  output file
    Do I = 1,MD_Steps

       Do Idim=1,3
          Read(29,*)(HMatrix(Jdim,Idim),Jdim=1,3)
       End Do

       ! We do not distinguish between primitive and
       ! conventional yet, just write the castep cell.

       Write(31,'(a8,i5)')'PRIMVEC ',I
       Do Idim=1,3
          Write(31,200)(HMatrix(Jdim,Idim)*bohrtoang,Jdim=1,3)
       End Do
       Write(31,'(a8,i5)')'CONVVEC ',I      
       Do Idim=1,3
          Write(31,200)(HMatrix(Jdim,Idim)*bohrtoang,Jdim=1,3)
       End Do       

       Do J = 1,Num_Atoms
          Read(29,*)R(:,J)
          Read(29,*)tmp,tmp,tmp
       End Do

       Do J = 1,Num_Atoms
          Read(30,*)F(:,J)
       End Do

       !If (I==1) then
       !   R0unwrap = R
       !   Rdiff    = 0.0_dp
       !Else
       !   Rdiff    = R-R0unwrap
       !End If

       !If (I==1) then

       !   ! Calculate fractional co-ordinates
       !   Call Utility_CalcRFrac(HMatrix,Num_Atoms,R,Rfrac)

          ! Calculate wrapped fractional co-ordinates
       !   RFrac = RFrac - Anint(RFrac,kind=dp) + 0.5_dp

          ! Convert R to wrapped absolute co-ordinates
          ! i.e. all atoms inside the simulation cell
       !   Do J = 1,Num_Atoms
       !      R0wrap(:,J) = Matmul(HMatrix,RFrac(:,J))
       !   End Do
         
       !End If

       Write(31,'(a10,i5)')'PRIMCOORD ',I
       Write(31,'(2i5)')Num_Atoms,1
       Do J = 1,Num_Atoms
          !Write(31,100)ElementID(J),(R0wrap(:,J)+Rdiff(:,J))*BohrtoAng,F(:,J)*bohrtoang
          Write(31,100)ElementID(J),R(:,J)*BohrtoAng,F(:,J)*bohrtoang
       End Do


    End Do

    close(29)
    close(30)
    close(31)

  End Subroutine IO_WriteXSFFile

  Subroutine IO_WriteXmolFile

    !=========================================================================!
    ! Creates an Xmol trajectory run from the trajectory file which can be    !
    ! visualised using Gopenmol and/or several other freely available         !
    ! programs.                                                               !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !   ElementID, String with chemical symbol for element used.              !
    !   SeedName , ID string for this run taken from input filename.          !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    !   Global,                                                               !
    !   MD_Steps          , Number of output steps                            !
    !   Num_Atoms         , Number of atoms                                   !
    !   SeedName          , Calculation Seedname                              !
    !                                                                         !
    !   Utility,                                                              !
    !   CalcRFrac         , Calculates fractional co-ordinates                !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    !-------------------------------------------------------------------------!
    ! Written by David Quigley  Feb 2004                                      !
    !=========================================================================!
    Use Global    , only : Num_Atoms,MD_Steps,SeedName
    Use Utility   , only : Utility_CalcRFrac
    Implicit None

    ! Loop counters
    Integer(kind=int32) :: I,J,Idim,Jdim

    Integer(kind=int32) :: Itmp       ! Temporary integer place holder
    Integer(kind=int32) :: ierr       ! I/O error flag

    !Holders for co-ordinates
    Real(kind=dp),Allocatable,Dimension(:,:) :: R
    Real(kind=dp),Allocatable,Dimension(:,:) :: RFrac 

    Real(kind=dp),Allocatable,Dimension(:,:) :: R0unwrap,R0wrap,Rdiff

    Real(kind=dp),Dimension(1:3,1:3) :: HMatrix

    ! Filename for Xmol file
    Character(30) :: XmolFile

    ! Conversion from internal units to Angstroms
    Real(kind=dp),Parameter :: BohrToAng = 0.5292_dp

    Real(kind=dp) :: tmp

100 Format(A3,3F12.6)
200 Format("* ",9F12.6)
300 Format(I4)

    Allocate(R(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*) "Error in IO_WriteXmolFile - Could not allocate R"
       Stop
    end if
    Allocate(RFrac(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*)"O_WriteXmolFile - Could not allocate RFrac"
       Stop
    end if
    Allocate(R0unwrap(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*)"O_WriteXmolFile - Could not allocate R0unwrap"
       Stop
    end if
    Allocate(R0wrap(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*)"O_WriteXmolFile - Could not allocate R0wrap"
       Stop
    end if
    Allocate(Rdiff(1:3,1:Num_Atoms),stat=ierr)
    if (ierr/=0) then
       Write(*,*)"O_WriteXmolFile - Could not allocate Rdiff"
       Stop
    end if


    Open(unit=29,file="Production.out",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error opening Production.out file for reading."
       Stop
    End if

    ! Specify output filename
    XMolFile  = trim(SeedName)//'.xmol' 

    Open(30,file=trim(XMolFile),status='replace',iostat=ierr)
    If (ierr.NE.0) then
       Write(*,*)'Error opening '//xmolfile//'for output'
       Stop
    End If

    ! Skip past header
    Read(29,'(2i15)' )Num_Atoms,Itmp  ! Number of atoms,Dimensionality
    Read(29,'(i15)'  )MD_Steps        ! Number of outputs in file
    Read(29,'(i15)'  )Itmp            ! Interval in timesteps
    Read(29,'(f15.5)')tmp             ! Calculation Timestep

    ! Create the Xmol  output file
    Do I = 1,MD_Steps

       Do Idim=1,3
          Read(29,*)(HMatrix(Jdim,Idim),Jdim=1,3)
       End Do
       Write(30,300)Num_Atoms
       Write(30,200)(HMatrix(:,Jdim)*BohrtoAng,Jdim=1,3)
       
       Do J = 1,Num_Atoms
          Read(29,*)R(:,J)
          Read(29,*)tmp,tmp,tmp
       End Do

       If (I==1) then
          R0unwrap = R
          Rdiff    = 0.0_dp
       Else
          Rdiff    = R-R0unwrap
       End If

       !If (I==1) then

          ! Calculate fractional co-ordinates
          !Call Utility_CalcRFrac(HMatrix,Num_Atoms,R,Rfrac)

          ! Calculate wrapped fractional co-ordinates
          ! RFrac = RFrac - Anint(RFrac,kind=dp) + 0.5_dp

          ! Convert R to wrapped absolute co-ordinates
          ! i.e. all atoms inside the simulation cell
          !Do J = 1,Num_Atoms
          !   R0wrap(:,J) = Matmul(HMatrix,RFrac(:,J))
          !End Do

         
       !End If

       Do J = 1,Num_Atoms
          !Write(30,100)ElementID(J),(R0wrap(:,J)+Rdiff(:,J))*BohrtoAng
          Write(30,100)ElementID(J),R(:,J)*BohrtoAng
          !Write(30,100)ElementID(J),R0wrap(:,J)*BohrtoAng
       End Do


    End Do

    Close(30)
    Close(29)

  End Subroutine IO_WriteXmolFile

End Module IO
