! -*- mode: F90 ; mode: font-lock ; column-number-mode: true ; vc-back-end: RCS -*-
!=============================================================================!
!                          U T I L I T Y                                      !
!=============================================================================!
!                                                                             !
! $Id: Utility.F90,v 1.1.1.1 2004/01/29 16:31:14 dq100 Exp $
!                                                                             !
!-----------------------------------------------------------------------------!
! Contains various routines for translating between fractional and absolute   !
! co-ordinates, wrappers for matrix diagonalisers etc.
!-----------------------------------------------------------------------------!
! Written by D. Quigley  - University of York                                 !
!-----------------------------------------------------------------------------!
!                                                                             !
! $Log: Utility.F90,v $
! Revision 1.1.1.1  2004/01/29 16:31:14  dq100
! Initial revision
!
!
!                                                                             !
!=============================================================================!


Module Utility
  Use constants, Only : dp,int32                !Minimal useage where practical

  Implicit None                                 !Impose strong typing
  Private                                       !Everything is private ...

  !---------------------------------------------------------------------------!
  !                       P u b l i c   R o u t i n e s                       !
  !---------------------------------------------------------------------------!
  Public :: Utility_CalcRFrac                  ! Calculate fractional co-ords
  Public :: Utility_Determinant                ! Determinant of a matrix

  !---------------------------------------------------------------------------!
  !                        P u b l i c   V a r i a b l e s                    !
  !---------------------------------------------------------------------------!

  !---------------------------------------------------------------------------!
  !                      P r i v a t e   R o u t i n e s                      !
  !---------------------------------------------------------------------------!


  !---------------------------------------------------------------------------!
  !                      P r i v a t e   V a r i a b l e s                    !
  !---------------------------------------------------------------------------!



Contains


  Subroutine Utility_CalcRFrac(HMatrix,Num_Atoms,Runwrap,Rfrac)

    !=========================================================================!
    ! Routine to calculate the fractional co-ordinates of the system from the !
    ! current absolute co-ordinates and the matrix of cell vectors.           !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    ! HMatrix      , Intent IN     , Matrix of Cell Vectors                   !
    ! Num_Atoms    , Intent IN     , Number of atoms in cell                  !
    ! Runwrap      , Intent IN     , Matrix of unwrapped co-ords              !
    ! Rfrac        , Intent OUT    , Matrix of fractional co-ords             !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !                                                                         !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! Arguments must be valid.                                                !
    !-------------------------------------------------------------------------!
    ! Written by David Quigley June 2003                                      !
    !=========================================================================!
    Implicit None

    ! Matrix of cell vectors
    Real(kind=dp),Dimension(3,3),Intent(IN) :: HMatrix

    ! Number of atoms
    Integer(kind=int32),Intent(IN) :: Num_Atoms

    ! Co-ord arrays
    Real(kind=dp),Dimension(3,Num_Atoms),Intent (IN) :: Runwrap
    Real(kind=dp),Dimension(3,Num_Atoms),Intent(OUT) :: Rfrac
    
    ! Loop counters
    Integer(kind=int32) :: I

    ! Local Cache variables
    Real(kind=dp)       :: tmpCache
    Real(kind=dp) :: XcoeffA,YcoeffA,ZcoeffA
    Real(kind=dp) :: XcoeffB,YcoeffB,ZcoeffB
    Real(kind=dp) :: XcoeffC,YcoeffC,ZcoeffC

    Real(kind=dp),Parameter :: one = 1.0_dp


    ! 3D system
    tmpCache = -HMatrix(3,1)*HMatrix(2,2)*HMatrix(1,3) &
         -HMatrix(2,1)*HMatrix(1,2)*HMatrix(3,3) &
         -HMatrix(1,1)*HMatrix(3,2)*HMatrix(2,3) &
         +HMatrix(3,1)*HMatrix(1,2)*HMatrix(2,3) &
         +HMatrix(1,1)*HMatrix(2,2)*HMatrix(3,3) &
         +HMatrix(2,1)*HMatrix(3,2)*HMatrix(1,3)
    tmpCache = one/tmpCache

    ! Calculate the A coefficients
    XcoeffA = HMatrix(2,2)*HMatrix(3,3) - HMatrix(3,2)*HMatrix(2,3)
    YcoeffA = HMatrix(3,2)*HMatrix(1,3) - HMatrix(1,3)*HMatrix(3,3)
    ZcoeffA = HMatrix(1,2)*HMatrix(2,3) - HMatrix(2,2)*HMatrix(1,3)

    ! Calculate the B coefficients
    XcoeffB = HMatrix(2,1)*HMatrix(3,3) - HMatrix(3,1)*HMatrix(2,3)
    YcoeffB = HMatrix(3,1)*HMatrix(1,3) - HMatrix(1,1)*HMatrix(3,3)
    ZcoeffB = HMatrix(1,1)*HMatrix(2,3) - HMatrix(2,1)*HMatrix(1,3)

    ! Calculate the C coefficients
    XcoeffC = HMatrix(2,1)*HMatrix(3,2) - HMatrix(3,1)*HMatrix(2,2)
    YcoeffC = HMatrix(3,1)*HMatrix(1,2) - HMatrix(1,1)*HMatrix(3,2)
    ZcoeffC = HMatrix(1,1)*HMatrix(2,2) - HMatrix(2,1)*HMatrix(1,2)

    ! Calculate the fractional co-ordinates
    RFrac(1,:) =  tmpCache*(Runwrap(1,:)*XcoeffA + Runwrap(2,:)*YcoeffA & 
         + Runwrap(3,:)*ZCoeffA)
    RFrac(2,:) = -tmpCache*(Runwrap(1,:)*XcoeffB + Runwrap(2,:)*YcoeffB &
         + Runwrap(3,:)*ZCoeffB)
    RFrac(3,:) =  tmpCache*(Runwrap(1,:)*XcoeffC + Runwrap(2,:)*YcoeffC &
         + Runwrap(3,:)*ZcoeffC)

    Return

  End Subroutine Utility_CalcRFrac

  Subroutine Utility_Determinant(Dim,Matrix,Det)

    !=========================================================================!
    ! Calculates the determinant of a 1x1, 2x2, or 3x3 Matrix                 !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !   Dim   , Dimensionality of Matrix passed in                            !
    !   Matrix, Matrix that we want the determinant of                        !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! Argument must be valid. Dim must equal 1,2 or 3                         !
    !-------------------------------------------------------------------------!
    ! Written by David Quigley June 2003                                      !
    !=========================================================================!

    Implicit None
    Integer(kind=int32),Intent(IN) :: Dim
    Real(kind=dp),Dimension(1:Dim,1:Dim),Intent(IN) :: Matrix

    Real(kind=dp),Intent(OUT) :: Det

    Select Case (Dim)
       Case (1)
          Det = Matrix(1,1)
       Case (2)
          Det = Matrix(1,1)*Matrix(2,2)-Matrix(1,2)*Matrix(2,1)
       Case (3)
          Det =       Matrix(1,1)*(Matrix(2,2)*Matrix(3,3) - &
                      Matrix(2,3)*Matrix(3,2))
          Det = Det - Matrix(1,2)*(Matrix(2,1)*Matrix(3,3) - &
                      Matrix(2,3)*Matrix(3,1))

          Det = Det + Matrix(1,3)*(Matrix(2,1)*Matrix(3,2) - &
                      Matrix(2,2)*Matrix(3,1))
    End Select

    Return

  End Subroutine Utility_Determinant

End Module Utility
