! -*- mode: F90 ; mode: font-lock ; column-number-mode: true ; vc-back-end: RCS -*-
!=============================================================================!
!                          A N A L Y S E                                      !
!=============================================================================!
!                                                                             !
! $Id: analyse.F90,v 1.5 2004/07/06 09:07:17 dq100 Exp $
!                                                                             !
!-----------------------------------------------------------------------------!
! This module performs calculations on MD trajectories located in the files   !
! Production.out, Sample.dat, Energy.dat and Cell.dat. These are generated    !
! from CASTEP .md files using routines in the IO.Module.                      !
!                                                                             !
!-----------------------------------------------------------------------------!
! Written by D. Quigley  - University of York                                 !
!-----------------------------------------------------------------------------!
!                                                                             !
! $Log: analyse.F90,v $
! Revision 1.5  2004/07/06 09:07:17  dq100
! Changed volume fluctional IO to GPa
!
! Revision 1.4  2004/03/08 18:41:24  dq100
! corrected pbc bug
!
! Revision 1.3  2004/02/02 15:27:04  dq100
! First stab at correcting for possible centre of mass motion. Currently assumes
! all species are of the same mass -- not good
!
! Revision 1.2  2004/01/30 10:02:55  dq100
! Corrected drop out from Analyse_AlphaP
!
! Revision 1.1.1.1  2004/01/29 16:31:14  dq100
! Initial revision
!
!
!                                                                             !
!=============================================================================!


Module analyse
  Use constants, Only : dp,int32                !Minimal useage where practical

  Implicit None                                 !Impose strong typing
  Private                                       !Everything is private ...

  !---------------------------------------------------------------------------!
  !                       P u b l i c   R o u t i n e s                       !
  !---------------------------------------------------------------------------!
  Public :: Analyse_CalcRDF                     !Radial distribution function
  Public :: Analyse_VACFunction                 !Vel autocorrelation function
  Public :: Analyse_MSD                         !Mean square displacement

  Public :: Analyse_HeatCap                     !Heat capacity from flucts
  Public :: Analyse_AlphaP                      !Expansion co-efficient (NPT only)
  Public :: Analyse_BetaT                       !Compressibility from flucts

  Public :: Analyse_TempTest                    !Analyse temperature distrib
  Public :: Analyse_VolTest                     !Analyse volume distrib

! Public :: Analyse_CellAngles                  !Calc cell angles/lengths

  ! Plots correlation function for temperature volume and pressure.
! Public :: Analyse_Correlations               

  !---------------------------------------------------------------------------!
  !                        P u b l i c   V a r i a b l e s                    !
  !---------------------------------------------------------------------------!

  !---------------------------------------------------------------------------!
  !                      P r i v a t e   R o u t i n e s                      !
  !---------------------------------------------------------------------------!


  !---------------------------------------------------------------------------!
  !                      P r i v a t e   V a r i a b l e s                    !
  !---------------------------------------------------------------------------!

  Real(kind=dp),save :: Analyse_LocalBT=-1.0_dp   ! Beta_T in a.u.


Contains

  Subroutine Analyse_CalcRDF(FirstStep,LastStep)
    !=========================================================================!
    ! This routine calculates a radial distribution function averaged over    !
    ! all steps between the two arguments given (inclusive).                  !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !	FirstStep, Intent IN  , Timestep at which averaging is to begin       !
    !   LastStep , Intent IN  , Timestep at which averaging is to end         !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !                Num_Atoms   , Number of atoms in the output              !
    !                seedname    , name for the castep run under analysis     !
    !                                                                         !
    ! Utility     ,  CalcRFrac   , Calculates fractional co-ordinates         !
    !                Determinant , Calculates determinants of matrices        !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! A Production.out file should exist in the current directory             !
    !                                                                         !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!
    Use Global  , Only : MD_Steps, Num_Atoms, seedname
    Use Utility , Only : Utility_CalcRFrac,Utility_Determinant
    Implicit None

    ! Number of spherical RDF shells
    Integer(kind=int32),Parameter :: NumShells=128
    Real   (kind=dp)   ,Parameter :: FourPI=12.56637061_dp
    Real   (kind=dp)   ,Parameter :: half = 0.5_dp

    ! Input parameters
    Integer(kind=int32),Intent(IN) :: FirstStep
    Integer(kind=int32),Intent(IN) :: LastStep 

    ! Unwrapped and fractional arrays
    Real(kind=dp),Dimension(3,Num_Atoms) :: Runwrap , Rfrac

    ! RDF arrays
    Real   (kind=dp)               ,Dimension(NumShells) :: RDF
    Integer(kind=int32),Allocatable,Dimension(:,:)       :: intRDF

    ! Matrix of cellvectors
    Real(kind=dp),Dimension(3,3) :: HMatrix


    ! PBC arrays
    Real(kind=dp),Dimension(3)      :: Min_rij
    Real(kind=dp)                   :: Min_Length
    Real(kind=dp),Dimension(3,8)    :: TransVec
    Real(kind=dp),Dimension(8)      :: Length_Rij
    Real(kind=dp),Dimension(3) :: Rij,Rijfrac

    Real(kind=dp) :: DR,tmp,RSquare,Maxrange,rho
    Real(kind=dp) :: DeltaT,factor,volume,R

    ! Variables required for BLAS call
    Character(1) :: OP='N'
    Real(kind=dp),Dimension(3) :: Y

    Integer(kind=dp) :: FileOutputInt,DummyInt

    ! Counters
    Integer(kind=int32)  :: I,J,K,L,RDFcount,Idim,Jdim,Icell,Jcell,trans


    ! Error flag
    Integer(kind=int32) :: ierr


    ! Checks
    If (LastStep>MD_Steps) then
       Write(*,*)"Error in Analyse_CalcRDF!"
       Write(*,*)"Final sample point is beyond end of run."
       Stop
    End If
    If (FirstStep<0.OR.LastStep>MD_Steps) then
       Write(*,*)"Error in Analyse_CalcRDF!"
       Write(*,*)"Start point is invalid."
       Stop
    End If
    If (FirstStep>LastStep) then
       Write(*,*)"Error in Analyse_CalcRDF!"
       Write(*,*)"End pos is before start pos."
       Stop
    End If

    ! Allocate arrays
    Allocate(intRDF(1:Num_atoms,1:Numshells),stat=ierr)
    If (ierr/=0) Then
       Write(*,*)"Error in Analyse_CalcRDF!"
       Write(*,*)"Could not allocate intRDF"
       Stop
    End If

    ! Initialise
    intRDF   = 0
    RDF      = 0.0_dp
    RDFCount = 0

    Open(Unit=27,File='Production.out',status='old',iostat=ierr)
    If (ierr/=0) Then
       Write(*,*)"Error in Analyse_CalcRDF"
       Write(*,*)"Could not open Production.out file"
       Stop
    End If

    ! The maximum meaningfull range over which an RDF can
    ! be calculated is half the length of the shortest cell
    ! vector    
    MaxRange = 99999999.9_dp

    !Skip past header
    Read(27,'(2i15)' )Num_Atoms,DummyInt  ! Number of atoms
    Read(27,'(i15)'  )MD_Steps            ! Number of outputs in file
    Read(27,'(i15)'  )FileOutputInt       ! Interval in timesteps
    Read(27,'(f15.5)')DeltaT              ! Calculation Timestep

    Do I = 1,MD_Steps

       ! Read the matrix of cell vectors
       Read(27,*)HMatrix(:,1)
       Read(27,*)HMatrix(:,2)
       Read(27,*)HMatrix(:,3)

       Do J = 1,3

          tmp = 0.5_dp*SQRT(HMatrix(1,J)**2 + HMatrix(2,J)**2 + &
               HMatrix(3,J)**2)
          If (tmp<MaxRange) then
             MaxRange = tmp
          End If

       End Do

       ! Not bothered about trajectory for now
       Do J = 1,Num_Atoms
          Read(27,*)tmp,tmp,tmp
          Read(27,*)tmp,tmp,tmp
       End Do

    End Do

    ! Radial mesh spacing
    DR    = MaxRange/(Real(NumShells,kind=dp))

    Rewind(27)

    !Skip past header
    Read(27,'(2i15)' )Num_Atoms,DummyInt   ! Number of atoms
    Read(27,'(i15)'  )MD_Steps             ! Number of outputs in file
    Read(27,'(i15)'  )FileOutputInt        ! Interval in timesteps
    Read(27,'(f15.5)')DeltaT               ! Calculation Timestep

    !Skip over data until required time
    If (FirstStep.Gt.1) Then
       Do I=1,FirstStep-1
          Read(27,*)tmp,tmp,tmp
          Read(27,*)tmp,tmp,tmp
          Read(27,*)tmp,tmp,tmp
          Do J = 1,Num_atoms
             Read(27,*)tmp,tmp,tmp
             Read(27,*)tmp,tmp,tmp
          End Do
       End Do
    End If

    Do L = FirstStep,LastStep

       ! Read in data at required time 
       Read(27,*)HMatrix(:,1)
       Read(27,*)HMatrix(:,2)
       Read(27,*)HMatrix(:,3)
       Do I = 1,Num_Atoms
          Read(27,*)Runwrap(:,I)
          Read(27,*)tmp,tmp,tmp
       End Do

       intRDF= 0

       ! Calculate an array of fractional co-ordinates
       Call Utility_CalcRFrac(HMatrix,Num_Atoms,Runwrap,Rfrac)

       Do I = 1,Num_Atoms
          Do J = I+1,Num_Atoms   ! Loop over all atom pairs

             ! Obtain fractional vector separation
             RijFrac(:) = RFrac(:,I) - RFrac(:,J)

             ! And real separation.
             Rij = Runwrap(:,I) - Runwrap(:,J)

             ! Replace RijFrac with origin of cell containing Rij
             RijFrac = Real(Floor(RijFrac,kind=int32),kind=dp)

             ! And translate Rij into that cell.          
             do idim=1,3
                do jdim=1,3
                   rij(idim)=rij(idim)-rijfrac(jdim)*hmatrix(idim,jdim)
                end do
             end do

             !... and now calculate translation vectors from each
             !of eight corners (of parallelepiped that Contains Rij) to Rij ...
             Trans = 1
             TransVec(:,Trans) = rij(:)
             Trans = Trans + 1
             Do ICell = 1,3
                Transvec(:,Trans) = rij(:) - HMatrix(:,ICell)
                Trans = Trans + 1
             End Do
             Do ICell = 1,2
                Do JCell = ICell+1,3
                   TransVec(:,Trans) = rij(:) - HMatrix(:,ICell) - HMatrix(:,JCell)
                   Trans = Trans+1
                End Do
             End Do
             TransVec(:,Trans) = rij(:) - HMatrix(:,1) - HMatrix(:,2) &
                  - HMatrix(:,3)

             ! Calculate length
             length_rij = 0.0_dp
             Do Idim = 1,3
                length_rij(:) = length_rij(:) + transvec(Idim,:)**2
             End Do

             !... and finally chose the one With minimum length ...
             min_length = 9999999.9_dp
             Do Idim = 1,8
                ! Account for possible inconsistency in accuracy of positions
                ! and cell by only allowing distances greater than 1 Bohr.
                If (length_rij(Idim)<min_length) Then
                   min_length = length_rij(Idim)
                End If
             End Do

             If (min_length.Lt.MaxRange*MaxRange) Then        
                R = Sqrt(min_length)
                !Check which shell this radius is in and store         
                K = Int(R/DR) + 1
                If (k==1) then
                   Write(*,*)"WARNING in Analyse_CalcRDF!"
                   Write(*,*)"Detecting serious core overlap between atoms."
                End If
                intRDF(I,K) = intRDF(I,K) + 1
             End If

          End Do
       End Do

       !Add up the array
       !and normalise accordingly

       !Calculate RDF
       Call Utility_Determinant(3,HMatrix,volume)
       rho = Real(Num_atoms,kind=dp)/volume
       Factor = 1.0_dp/(half*Real(Num_atoms,kind=dp)*rho*FourPI*Dr)
       Do I = 1,NumShells

          R = (Real(I,kind=dp)-0.5_dp)*Dr
          RDF(I) = RDF(I) + Factor*Real(Sum(intRDF(:,I)),kind=dp)/(R**2)

       End Do

       RDFCount = RDFCount + 1

    End Do

    !Average the RDF
    RDF = RDF/Real(RDFCount,kind=dp)

    Close(27)

    Open(27,File=trim(seedname)//"_RDF.dat",status="replace",iostat=ierr)
   If (ierr/=0) then
       Write(*,*)"Error in Analyse_CalcRDF"
       Write(*,*)"Could not open "//trim(seedname)//"_RDF.dat"
       Stop
    End If

    Write(27,*)"# RDF file for castep MD run "//trim(seedname)
    Write(27,*)"# Calculated using mdtep."
    
    Do I = 1,NumShells
       Write(27,*)(Real(I,kind=dp)-0.5_dp)*DR,RDF(I)
    End Do

    Close(27)

    Deallocate(intRDF,stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_CalcRDF!"
       Write(*,*)"Could not deallocate radial bin array."
       Stop
    End If

    Return

  End Subroutine Analyse_CalcRDF

  Subroutine Analyse_VACFunction(Start,Final,Length,Interval)

    !=========================================================================!
    ! Calculates the time and atom averaged velocity autocorrelation function !
    ! between the start and end times specified in the arguments.             !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !	Start    , Intent IN  , Timestep at which averaging is to begin       !
    !   Final    , Intent IN  , Timestep at which averaging is to end         !
    !   Length   , Intent IN  , Length of correlation function                !
    !   Interval , Intent IN  , gap between time origins sampled              !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !                Num_Atoms   , Number of atoms in the output              !
    !                seedname    , name for the castep run under analysis     !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! A Production.dat file should exist in the current directory             !
    !                                                                         !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!
    Use Global   , Only : MD_Steps, Num_Atoms, seedname
    Implicit None
    Integer(kind=int32),Intent(IN)  :: Start,Final,Length,Interval

    Real(kind=dp),Allocatable,Dimension(:)     :: VACTotal
    Real(kind=dp),Allocatable,Dimension(:)     :: VAC
    Real(kind=dp),Allocatable,Dimension(:,:,:) :: V
    Real(kind=dp)  :: tmp,Norm,invN,deltaT

    Real(kind=dp),Dimension(3) :: Vcom

    Integer(kind=int32)  :: StartPos,EndPos,I,J,M !Loop counters/markers
    Integer(kind=int32)  :: ierr,FileOutputInt,DummyInt


    ! NEEDS MODIFYING TO TAKE OUT C.O.M. Motions

    ! Checks
    If (Final>MD_Steps) then
       Write(*,*)"Error in Analyse_VACFunction!"
       Write(*,*)"Final sample point is beyond end of run."
       Stop
    End If
    If (Start<0.OR.Start>MD_Steps) then
       Write(*,*)"Error in Analyse_VACFunction!"
       Write(*,*)"Start point is invalid."
       Stop
    End If
    If (Start>Final) then
       Write(*,*)"Error in Analyse_VACFunction!"
       Write(*,*)"End pos is before start pos."
       Stop
    End If

    invN = 1.0_dp/Real(Num_atoms,kind=dp)

    !Allocate heap space for V
    Allocate(V(3,Num_atoms,Length),stat=ierr)
    If (ierr/=0) then
       Print*,"Could not allocate velocity array"
       Stop
    End If
    Allocate(VAC(1:Length),stat=ierr)
    If (ierr/=0) then
       Print*,"Could not allocate VAC array"
       Stop
    End If

    Allocate(VACTotal(1:Length),stat=ierr)
    If (ierr/=0) then
       Print*,"Could not allocate VACTOTAL array"
       Stop
    End If

    !Initialise Start and End positions
    StartPos = Start
    EndPos   = Start + Length

    !Initialise VAC arrays and number of VACS
    VAC = 0.0_dp
    VACTotal = 0.0_dp
    M = 0

    !Open the results file
    Open(Unit=27,File="Production.out",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_VACFunction!"
       Write(*,*)"Could not open Production.out file."
       Stop
    End If

    !Skip past header information
    Read(27,'(2i15)' )Num_Atoms,Dummyint   ! Number of atoms
    Read(27,'(i15)'  )MD_Steps             ! Number of outputs in file
    Read(27,'(i15)'  )FileOutputInt        ! Interval in timesteps
    Read(27,'(f15.5)')DeltaT               ! Calculation Timestep

    !Skip past unwanted data
    Do I = 1,Start -1
       Read(27,*)tmp,tmp,tmp
       Read(27,*)tmp,tmp,tmp
       Read(27,*)tmp,tmp,tmp        
       Do J = 1,Num_atoms
          Read(27,*)tmp,tmp,tmp
          Read(27,*)tmp,tmp,tmp
       End Do
    End Do

    !Read data for 1st VAC
    Do I = 1,Length
       Read(27,*)tmp,tmp,tmp
       Read(27,*)tmp,tmp,tmp
       Read(27,*)tmp,tmp,tmp
       Vcom(:) = 0.0_dp
       Do J = 1,Num_atoms
          Read(27,*)tmp,tmp,tmp
          Read(27,*)V(1,J,I),V(2,J,I),V(3,J,I)
          Vcom(:) = Vcom(:) + V(:,J,I)
       End Do

       ! We now calculate the C.O.M. velocity
       ! for now assumes equal masses
       Vcom(:) = Vcom(:)/Real(Num_Atoms,kind=dp)

       ! And subtract
       Do J = 1,Num_atoms
          V(:,J,I) = V(:,J,I) - Vcom(:)
       End Do

    End Do

    Do   !Loop over calculated VACs

       !Initialise at first point in this VAC
       VAC = 0.0_dp

       !Make 1st point in this VAC and store V0
       Do J=1,Num_atoms
          VAC(1) = VAC(1) + Dot_product(V(:,J,1),V(:,J,1))
       End Do

       Do I = 2,Length   !Loop over points in the VAC

          Do J = 1,Num_Atoms     !Loop over atoms in the ensemble
           
             !Read Data and calculate VAC
             VAC(I) = VAC(I) + Dot_product(V(:,J,I),V(:,J,1))

          End Do

       End Do

       !Accumate the sum of all normalised VACs so far
       Norm = 1.0_dp/VAC(1)
       VACTotal = VACTotal+(VAC*Norm)

       !Increment start and End positions of VAC
       StartPos = StartPos + Interval
       EndPos   = EndPos + Interval

       !Increment no of calculated VACS
       M = M +1

       !Get out of loop if no more VACs can be generated
       If (EndPos>Final) Then
          Exit
       End If

       !Adjust array to hold data for next VAC
       Do I = 1,Length-Interval
          Do J = 1,Num_atoms
             V(1,J,I) = V(1,J,I+Interval)
             V(2,J,I) = V(2,J,I+Interval)
             V(3,J,I) = V(3,J,I+Interval)
          End Do
       End Do

       !Complete array with new data
       Do I = Length-Interval+1,Length
          Read(27,*)tmp,tmp,tmp
          Read(27,*)tmp,tmp,tmp
          Read(27,*)tmp,tmp,tmp
          Vcom = 0.0_dp
          Do J = 1,Num_Atoms
             Read(27,*)tmp,tmp,tmp
             Read(27,*)V(1,J,I),V(2,J,I),V(3,J,I)
             Vcom(:) = Vcom(:) + V(:,J,I)
          End Do
       
          

          ! We now calculate the C.O.M. velocity
          ! for now assumes equal masses
          Vcom(:) = Vcom(:)/Real(Num_Atoms,kind=dp)

          ! And subtract
          Do J = 1,Num_atoms
             V(:,J,I) = V(:,J,I) - Vcom(:)
          End Do
       

       End Do

    End Do

    !Calculate averaged VAC
    VACTotal = VACTotal/Real(M,kind=dp)


    Close(27)

    Open(27,File=trim(seedname)//"_VAC.dat",status='replace',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_VACFunction"
       Write(*,*)"Could not open "//trim(seedname)//"_VAC.dat"
       Stop
    End If

    Write(27,*)"# Velocity autocorrelation function for CASTEP run "//trim(Seedname)
    Write(27,*)"# Calculated using mdtep."

    Do I = 1,Length
       Write(27,*)Real(I-1,kind=dp)*DeltaT*Fileoutputint*0.00002417_dp,VACTOTAL(I)
    End Do

    Close(27)

    !Close the results file and release memory  
    Deallocate(V,stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_VACFunction!"
       Write(*,*)"Could not deallocate V"
       Stop
    End If
    Deallocate(VAC,stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_VACFunction!"
       Write(*,*)"Could not deallocate VAC"
       Stop
    End If
    Deallocate(VACTotal,stat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_VACFunction!"
       Write(*,*)"Could not deallocate VACTOTAL"
       Stop
    End If

  End Subroutine Analyse_VACFunction

  Subroutine Analyse_MSD(Start,Final,Length,Interval)
    !=========================================================================!
    ! Calculates the Mean squared displacement of atoms averaged over all     !
    ! atoms and over time origins as specified by the arguments               !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !	Start    , Intent IN  , Timestep at which averaging is to begin       !
    !   Final    , Intent IN  , Timestep at which averaging is to end         !
    !   Length   , Intent IN  , Length of MSD function                        !
    !   Interval , Intent IN  , gap between time origins sampled              !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !                Num_Atoms   , Number of atoms in the output              !
    !                seedname    , name for the castep run under analysis     !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! A Production.out file should exist in the current directory             !
    !                                                                         !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!
    Use Global  , Only : MD_Steps, Num_Atoms, seedname
    Implicit None

    Integer(kind=int32),Intent(IN) :: Start
    Integer(kind=int32),Intent(IN) :: Final
    Integer(kind=int32),Intent(IN) :: Length
    Integer(kind=int32),Intent(IN) :: Interval

    Real(kind=dp),Dimension(Length)           :: MSD
    Real(kind=dp),Dimension(Num_Atoms,Length) :: X,Y,Z

    Integer(kind=int32) :: I,J,ierr,fileoutputint,DummyInt
    Integer(kind=int32) :: NumMSDs = 0

    ! Position of centre of mass
    Real(kind=dp) :: Xcom,Ycom,Zcom

    Real(kind=dp) :: totdisp,sqdisp,tmp,deltat,avx,avy,avz

    MSD = 0.0_dp

    ! NEEDS MODIFYING TO TAKE OUT C.O.M. Motions

    ! Checks
    If (Final>MD_Steps) then
       Write(*,*)"Error in Analyse_MSD!"
       Write(*,*)"Final sample point is beyond end of run."
       Stop
    End If
    If (Start<0.OR.Start>MD_Steps) then
       Write(*,*)"Error in Analyse_MSD!"
       Write(*,*)"Start point is invalid."
       Stop
    End If
    If (Start>Final) then
       Write(*,*)"Error in Analyse_MSD!"
       Write(*,*)"End pos is before start pos."
       Stop
    End If

    !Open the output file
    Open(Unit=27,File="Production.out",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_MSD"
       Write(*,*)"Could not open Production.out file."
       Stop
    End If

    ! Skip past the header
    Read(27,'(2i15)' )Num_Atoms,DummyInt     ! Number of atoms
    Read(27,'(i15)'  )MD_Steps               ! Number of outputs in file
    Read(27,'(i15)'  )FileOutputInt          ! Interval in timesteps
    Read(27,'(f15.5)')DeltaT                 ! Calculation Timestep

    !Skip past unwanted data
    Do I = 1,Start -1
       Read(27,*)tmp,tmp,tmp
       Read(27,*)tmp,tmp,tmp
       Read(27,*)tmp,tmp,tmp        
       Do J = 1,Num_atoms
          Read(27,*)tmp,tmp,tmp
          Read(27,*)tmp,tmp,tmp
       End Do
    End Do

    ! Read in data for first MSD
    Do I = 1,Length
       !Read in the current cell dimension
       Read(27,*,err=10,End=10)tmp,tmp,tmp
       Read(27,*,err=10,End=10)tmp,tmp,tmp
       Read(27,*,err=10,End=10)tmp,tmp,tmp

       !Read in the positions and find rcom
       Xcom = 0.0_dp
       Ycom = 0.0_dp
       Zcom = 0.0_dp
       Do J = 1,Num_atoms
          Read(27,*,err=10,End=10)X(J,I),Y(J,I),Z(J,I)
          Read(27,*,err=10,End=10)tmp,tmp,tmp
          Xcom = Xcom + X(J,I)
          Ycom = Ycom + Y(J,I)
          Zcom = Zcom + Z(J,I)
       End Do

       Xcom = Xcom/Real(Num_Atoms,kind=dp)
       Ycom = Ycom/Real(Num_Atoms,kind=dp)
       Zcom = Zcom/Real(Num_Atoms,kind=dp)
       
       X(:,I) = X(:,I) - Xcom
       Y(:,I) = Y(:,I) - Ycom       
       Z(:,I) = Z(:,I) - Zcom     

    End Do

    Do   ! Loop over time origins

       Do I = 2, Length !Loop over all other snapshots

          totdisp = 0.0_dp

          Do J=1,Num_atoms

             ! Co-ords in the Production.out file are always
             ! unwrapped for castep runs. - No boundary
             ! condition hastles.

             ! For each atom calculate the Squared displacement
             sqdisp = (X(J,I)-X(J,1))**2 + (Y(J,I)-Y(J,1))**2 & 
                  + (Z(J,I)-Z(J,1))**2
             totdisp = totdisp + sqdisp

          End Do

          !Calculate the mean square displacement
          MSD(I) = MSD(I) + totdisp/Real(Num_atoms,kind=dp)

       End Do

       NumMSDs = NumMSDs + 1

       ! Shuffle
       Do I = 1,Length - Interval
          Do J = 1,Num_atoms
             X(J,I) = X(J,I+Interval)
             Y(J,I) = Y(J,I+Interval)
             Z(J,I) = Z(J,I+Interval)
          End Do
       End Do

       ! Read in next point or skip to end if no more points
       Do I = Length-Interval+1,Length

          !Read in the current cell dimension
          Read(27,*,err=10,End=10)tmp,tmp,tmp
          Read(27,*,err=10,End=10)tmp,tmp,tmp
          Read(27,*,err=10,End=10)tmp,tmp,tmp

          !Read in the positions
          Xcom = 0.0_dp
          Ycom = 0.0_dp
          Zcom = 0.0_dp
          Do J = 1,Num_atoms
             Read(27,*,err=10,End=10)X(J,I),Y(J,I),Z(J,I)
             Read(27,*,err=10,End=10)tmp,tmp,tmp
             Xcom = Xcom + X(J,Length)
             Ycom = Ycom + Y(J,Length)
             Zcom = Zcom + Z(J,Length)
          End Do

          Xcom = Xcom/Real(Num_Atoms,kind=dp)
          Ycom = Ycom/Real(Num_Atoms,kind=dp)
          Zcom = Zcom/Real(Num_Atoms,kind=dp)
       
          X(:,I) = X(:,I) - Xcom
          Y(:,I) = Y(:,I) - Ycom       
          Z(:,I) = Z(:,I) - Zcom  

       End Do

    End Do


10  MSD = MSD/Real(NumMSDs,kind=dp)


    Close(27)

    Open(27,File=trim(seedname)//"_MSD.dat",status='replace',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_MSD!"
       Write(*,*)"Could not open "//trim(seedname)//"_MSD.dat for output"
       Stop
    End If

    Write(27,*)"# Mean square displacement for CASTEP run "//trim(Seedname)
    Write(27,*)"# Calculated using mdtep."

    Do I = 1,Length
       Write(27,*)Real(I,kind=dp)*DeltaT*FileOutputInt*0.00002417,MSD(I)
    End Do
    Close(27)

  End Subroutine Analyse_MSD

  Subroutine Analyse_HeatCap(Start,Final)
    !=========================================================================!
    ! Calculates heat capacity C_v for NVT runs, and C_p for NPT runs from    !
    ! fluctuation in energy and volume. Also performs a blocking analysis     !
    ! and writes the results to file for plotting and error estimation.       !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !	Start    , Intent IN  , Timestep at which averaging is to begin       !
    !   Final    , Intent IN  , Timestep at which averaging is to end         !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !                Num_Atoms   , Number of atoms in the output              !
    !                seedname    , name for the castep run under analysis     !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! A  Sample.dat file should exist in the current directory                !
    ! A  Energy.dat  "     "      "    "  "     "        "                    !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!
    Use Global  , Only  : MD_Steps, Num_Atoms, seedname
    Use Constants, Only : kb
    Implicit None

    Integer(kind=int32),Intent(IN) :: Start,Final 

    Integer(kind=int32) :: CorrCount,CorrNSamp,N,Linecount,I,ierr

    Real(kind=dp) :: Temp,Volume,TMean,EnthalpMean,Tsd,Enthalpsd
    Real(kind=dp) :: Enthalpy,Pressure,MeanP,totalE
    Real(kind=dp) :: Terr,Enthalperr,Tmeansq,Enthalpmeansq
    Real(kind=dp) :: dummy
    Real(kind=dp) :: Cp,Cperr,VvarErr
    Real(kind=dp) :: BlockTemp,BlockEnthalpy
    Real(kind=dp) :: Enthalpmeansqsq,ErrMeansqEn

    Logical :: Firstpass = .TRUE.

    ! Checks
    If (Final>MD_Steps) then
       Write(*,*)"Error in Analyse_HeatCap!"
       Write(*,*)"Final sample point is beyond end of run."
       Stop
    End If
    If (Start<0.OR.Start>MD_Steps) then
       Write(*,*)"Error in Analyse_HeatCap!"
       Write(*,*)"Start point is invalid."
       Stop
    End If
    If (Start>Final) then
       Write(*,*)"Error in Analyse_HeatCap!"
       Write(*,*)"End pos is before start pos."
       Stop
    End If


    ! Number of samples
    N = Final-Start

    ! Open the Sample.dat file
    Open(unit=23,file="Sample.dat",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_HeatCap!"  
       Write(*,*)"Could not open Sample.dat file."
       Stop
    End If
    ! Open the Energy.dat file
    Open(unit=24,file="Energy.dat",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_HeatCap!"  
       Write(*,*)"Could not open Energy.dat file."
       Stop
    End If


    ! Open the Output file
    Open(27,File=trim(seedname)//"_heatcap.dat",status='replace',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_HeatCap!"
       Write(*,*)"Could not open "//trim(seedname)//"_heatcap.dat for output"
       Stop
    End If

    Write(27,*)"# Heat Capacity for CASTEP run "//trim(Seedname)
    Write(27,*)"# Calculated using mdtep."

    ! Initial block size is one to obtain the total
    ! averages - used to calculate central Cp 
    ! value
    CorrNSamp = 0

    ! We need the average pressure to continure
    MeanP = 0.0_dp

    ! Loop over unwanted data
    Do I = 1,Start-1
       Read(23,*)Dummy,Dummy,Pressure,Dummy
    End Do


    Do I = Start,Final
       Read(23,*)Dummy,Dummy,Pressure,Dummy
       MeanP = MeanP + Pressure
    End Do

    Rewind(23)

    ! Calculate average pressure
    MeanP = MeanP/Real(N,kind=dp)
    MeanP = MeanP/29414000.0_dp    ! Convert back into a.u.

    Do           ! Loop over block sizes 

       CorrNSamp = 1+CorrNSamp    ! increment block size 

       ! Rewind the input file from last block size
       Rewind(23)
       Rewind(24)

       ! Counter for reaching blocksize
       CorrCount = 0

       ! Zero all accumulators
       EnthalpMean     = 0.0_dp
       TMean           = 0.0_dp
       Tmeansq         = 0.0_dp
       Enthalpmeansq   = 0.0_dp
       Enthalpmeansqsq = 0.0_dp

       ! Zero accumulators within block
       BlockTemp   = 0.0_dp
       BlockEnthalpy = 0.0_dp

       N = 0

       ! Skip unwanted data
       Do I = 1,Start-1
          Read(23,*)dummy,temp,dummy,Volume
          Read(24,*)dummy,dummy,dummy,TotalE         
       End Do


       ! Loop over all entries in Sample.dat
       Do I = Start,Final
          ! Read in the temperature and volume
          Read(23,*)dummy,temp,dummy,Volume
          Read(24,*)dummy,dummy,dummy,TotalE

          ! Calculate Enthalpy
          Enthalpy = TotalE + MeanP*Volume

          ! Average within the current block
          CorrCount   = CorrCount + 1
          BlockTemp   = BlockTemp + Temp
          BlockEnthalpy = BlockEnthalpy + Enthalpy

          ! If we have reached the end of a block then
          ! accumumate the global averages
          If (CorrCount == CorrNSamp) then

             ! Average temp and volume for the preceeding block
             Temp = BlockTemp/Real(CorrNSamp,kind=dp)
             Enthalpy = BlockEnthalpy/Real(CorrNSamp,kind=dp)

             ! Zero the within-block averages
             BlockTemp = 0.0_dp
             BlockEnthalpy = 0.0_dp

             ! Accumunate the global means
             Tmean = Tmean + Temp
             EnthalpMean = EnthalpMean + Enthalpy

             ! And the global mean-squares
             Tmeansq = Tmeansq + Temp*Temp
             EnthalpMeansq = EnthalpMeansq + Enthalpy*Enthalpy

             ! And the mean-square-square of volume for error analysis
             EnthalpMeansqsq = EnthalpMeansqsq + Enthalpy**4

             ! Count the number of blocks so far
             N = N +1

             ! Reset the block position counter
             CorrCount =0

          End If

       End Do   ! End loop over entries in sample file



       ! If the number of blocked samples is less than
       ! 100 then get out.
       If (N.LT.5) then
          Write(27,*)
          Write(27,*)
          Write(27,*)"Less than 5 blocks - Stopping"
          Close(27)
          Close(23)
          Close(24)
          Return
       End If

       ! Calculate the mean and the error
       ! in the mean for both the temperature
       ! and volume in the usual way.

       Tmean = Tmean / Real(N,kind=dp)
       Tmeansq = Tmeansq / Real(N,kind=dp)

       Tsd  = SQRT(Tmeansq-Tmean*Tmean)
       Terr = Tsd/Real(N,kind=dp)

       EnthalpMean = Enthalpmean / Real(N,kind=dp)
       EnthalpMeansq = EnthalpMeansq / Real(N,kind=dp)

       EnthalpMeansqsq = EnthalpMeansqsq / Real(N,kind=dp)

       Enthalpsd = SQRT(EnthalpMeansq-Enthalpmean*Enthalpmean)
       Enthalperr = Enthalpsd/Real(N,kind=dp)

       ErrMeansqEn = SQRT(EnthalpMeanSqSQ - EnthalpMeansq*EnthalpMeansq)/Real(N,kind=dp)

       ! Calculate Cp -- only printed for block size of one.

       Cp = Enthalpsd*Enthalpsd/(Kb*TMean*Tmean)

       If (FirstPass) then 
          Write(27,*)
          Write(27,*)"Heat Capacity = ",Cp,"Hartree / Kelvin"
          Write(27,*)
          Write(27,'(2A20)')"Block Size","Error in C_{P}"
          Write(27,*)"========================================"
          FirstPass = .FALSE.
       End If

       ! Calculate error in Cp for the current blocksize
       CpErr =  ErrMeansqEn/(Kb*Tmean*Tmean) &
            -  2.0_dp*EnthalpMean*EnthalpErr/(Kb*Tmean*Tmean) &
            -  2.0_dp*Enthalpsd*Enthalpsd*Terr/(Kb*TMean*Tmean*Tmean)

       Write(27,'(I20,E20.8)')CorrNSamp,Cperr

    End Do ! End Loop over block sizes.

    Close(27)
    Close(23)
    Close(24)

    Return

  End Subroutine Analyse_HeatCap

  Subroutine Analyse_BetaT(Start,Final)

    !=========================================================================!
    ! Finds the Isothermal Bulk compressibility for NPT runs using the        !
    ! fluctuations in Temperature and Volume. Will return zero for a fixed    !
    ! cell calculation.                                                       !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !	Start    , Intent IN  , Timestep at which averaging is to begin       !
    !   Final    , Intent IN  , Timestep at which averaging is to end         !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    ! LocalBT , Used by the Volume distribution routine to fit Gaussian       !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !                Num_Atoms   , Number of atoms in the output              !
    !                seedname    , name for the castep run under analysis     !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! A  Sample.dat file should exist in the current directory                !
    ! A  Energy.dat  "     "      "    "  "     "        "                    !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!
    Use Global  , Only  : MD_Steps, Num_Atoms, seedname
    Use Constants, Only : kb
    Implicit None

    Integer(kind=int32),Intent(IN) :: Start,Final 

    Integer(kind=int32) :: CorrCount,CorrNSamp,N,Linecount,I,ierr

    Real(kind=dp) :: Temp,Volume,TMean,VolMean,Tsd,Volsd
    Real(kind=dp) :: Terr,Volerr,Tmeansq,Volmeansq
    Real(kind=dp) :: dummy
    Real(kind=dp) :: BetaT,BetaTerr,VvarErr
    Real(kind=dp) :: BlockTemp,BlockVolume,BlockVVarErr
    Real(kind=dp) :: Volmeansqsq,ErrMeansqVol

    Logical :: Firstpass = .TRUE.

    ! Checks
    If (Final>MD_Steps) then
       Write(*,*)"Error in Analyse_BetaT!"
       Write(*,*)"Final sample point is beyond end of run."
       Stop
    End If
    If (Start<0.OR.Start>MD_Steps) then
       Write(*,*)"Error in Analyse_BetaT!"
       Write(*,*)"Start point is invalid."
       Stop
    End If
    If (Start>Final) then
       Write(*,*)"Error in Analyse_BetaT!"
       Write(*,*)"End pos is before start pos."
       Stop
    End If


    ! Open the goldilocs Sample.dat file
    Open(unit=23,file="Sample.dat",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_BetaT!"  
       Write(*,*)"Could not open Sample.dat file."
       Stop
    End If

    ! Open the Output file
    Open(27,File=trim(seedname)//"_betat.dat",status='replace',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_BetaT!"
       Write(*,*)"Could not open "//trim(seedname)//"_betat.dat for output"
       Stop
    End If

    Write(27,*)"# Isothermal bulk compressibility for CASTEP run "//trim(Seedname)
    Write(27,*)"# Calculated using mdtep."


    ! Initial block size is one to obtain the total
    ! averages - used to calculate central Beta_{T} 
    ! value
    CorrNSamp = 0

    Do           ! Loop over block sizes 

       CorrNSamp = 1+CorrNSamp    ! increment block size 

       ! Rewind the input file from last block size
       Rewind(23)

       ! Counter for reaching blocksize
       CorrCount = 0

       ! Zero all accumulators
       VolMean     = 0.0_dp
       TMean       = 0.0_dp
       Tmeansq     = 0.0_dp
       Volmeansq   = 0.0_dp
       Volmeansqsq = 0.0_dp

       ! Zero accumulators within block
       BlockTemp   = 0.0_dp
       BlockVolume = 0.0_dp

       N = 0

       ! Skip past unwanted data
       Do I = 1,Start-1
          Read(23,*)dummy,Temp,dummy,Volume
       End Do

       ! Loop over all entries in Sample.dat
       Do I = Start,Final
          ! Read in the temperature and volume
          Read(23,*)dummy,Temp,dummy,Volume

          ! Average within the current block
          CorrCount   = CorrCount + 1
          BlockTemp   = BlockTemp + Temp
          BlockVolume = BlockVolume + Volume

          ! If we have reached the end of a block then
          ! accumumate the global averages
          If (CorrCount == CorrNSamp) then

             ! Average temp and volume for the preceeding block
             Temp = BlockTemp/Real(CorrNSamp,kind=dp)
             Volume = BlockVolume/Real(CorrNSamp,kind=dp)

             ! Zero the within-block averages
             BlockTemp = 0.0_dp
             BlockVolume = 0.0_dp

             ! Accumunate the global means
             Tmean = Tmean + Temp
             VolMean = VolMean + Volume

             ! And the global mean-squares
             Tmeansq = Tmeansq + Temp*Temp
             VolMeansq = VolMeansq + Volume*Volume

             ! And the mean-square-square of volume for error analysis
             VolMeansqsq = VolMeansqsq + Volume**4

             ! Count the number of blocks so far
             N = N +1

             ! Reset the block position counter
             CorrCount =0

          End If

       End Do   ! End loop over entries in sample file


       ! If the number of blocked samples is less than
       ! 100 then get out.
       If (N.LT.10) then
          Write(27,*)
          Write(27,*)
          Write(27,*)"Less than 10 blocks - Stopping"
          Close(23)
          Close(27)
          Return
       End If

       ! Calculate the mean and the error
       ! in the mean for both the temperature
       ! and volume in the usual way.

       Tmean = Tmean / Real(N,kind=dp)
       Tmeansq = Tmeansq / Real(N,kind=dp)

       Tsd  = SQRT(Tmeansq-Tmean*Tmean)
       Terr = Tsd/Real(N,kind=dp)

       VolMean = Volmean / Real(N,kind=dp)
       VolMeansq = VolMeansq / Real(N,kind=dp)

       VolMeansqsq = VolMeansqsq / Real(N,kind=dp)

       Volsd = SQRT(VolMeansq-volmean*volmean)
       Volerr = Volsd/Real(N,kind=dp)

       ErrMeansqVol = SQRT(VolMeanSqSQ - VolMeansq*VolMeansq)/Real(N,kind=dp)

       ! Calculate Bt -- only printed for block size of one.

       BetaT = Volsd*Volsd/(Volmean*Kb*TMean)

       ! Calculate error in Bt for the current blocksize
       BetaTerr = ErrMeanSqVol/(VolMean*Kb*Tmean) &
            -Volerr*VolMeanSq/(VolMean*VolMean*Kb*Tmean) &
            -Volerr/(Kb*Tmean) &
            - Terr*Volsd*Volsd/(VolMean*Kb*Tmean*Tmean)

       ! store as fractional error
       BetaTerr = BetaTerr/BetaT

       ! Convert BetaT to correct dimensions in units of GPa
       BetaT = 29414000.0_dp/(1000.0_dp*BetaT)

       If (FirstPass) then 
          Write(27,*)
          Write(27,*)"Beta_{T} = ",BetaT,"GPa"
          Write(27,*)
          Write(27,'(2A20)')"Block Size","Error in Beta_{T}"
          Write(27,*)"========================================"
          ! Set module level variable
          Analyse_LocalBT = BetaT
          FirstPass = .FALSE.
       End If

       Write(27,'(I20,E20.8)')CorrNSamp,BetaTerr*BetaT

    End Do ! End Loop over block sizes.


    Close(27)
    Close(23)
    Return

  End Subroutine Analyse_BetaT

  Subroutine Analyse_AlphaP(Start,Final)
    !=========================================================================!
    ! Finds the thermal expansion coefficient for NPT runs from the           !
    ! fluctuations in energy and temperature.                                 !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !	Start    , Intent IN  , Timestep at which averaging is to begin       !
    !   Final    , Intent IN  , Timestep at which averaging is to end         !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    ! LocalBT , Used by the Volume distribution routine to fit Gaussian       !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !                Num_Atoms   , Number of atoms in the output              !
    !                seedname    , name for the castep run under analysis     !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! A  Sample.dat file should exist in the current directory                !
    ! A  Energy.dat  "     "      "    "  "     "        "                    !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!
    Use Global  , Only  : MD_Steps, Num_Atoms, seedname
    Use Constants, Only : kb
    Implicit None

    Integer(kind=int32),Intent(IN) :: Start,Final 

    Integer :: CorrCount,CorrNSamp,N,Linecount,I,ierr

    Real(kind=dp) :: Temp,Volume,TMean,EnthalpMean,Tsd,Enthalpsd
    Real(kind=dp) :: Enthalpy,Pressure,MeanP,totalE
    Real(kind=dp) :: Terr,Enthalperr,Tmeansq,Enthalpmeansq
    Real(kind=dp) :: dummy
    Real(kind=dp) :: Cp,Cperr,VvarErr
    Real(kind=dp) :: BlockTemp,BlockEnthalpy
    Real(kind=dp) :: volmean,volmean2,volmeansq,volsd,volerr
    Real(kind=dp) :: Prod,prodmean,blockprod,prodsd,proderr,prodmeansq
    Real(kind=dp) :: Alpha_p,Alpha_perr
    Real(kind=dp) :: blockvol

    Logical :: Firstpass = .TRUE.

    ! Checks
    If (Final>MD_Steps) then
       Write(*,*)"Error in Analyse_AlphaP!"
       Write(*,*)"Final sample point is beyond end of run."
       Stop
    End If
    If (Start<0.OR.Start>MD_Steps) then
       Write(*,*)"Error in Analyse_AlphaP!"
       Write(*,*)"Start point is invalid."
       Stop
    End If
    If (Start>Final) then
       Write(*,*)"Error in Analyse_AlphaP!"
       Write(*,*)"End pos is before start pos."
       Stop
    End If


    ! Number of samples
    N = Final-Start

    ! Open the Sample.dat file
    Open(unit=23,file="Sample.dat",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_AlphaP!"  
       Write(*,*)"Could not open Sample.dat file."
       Stop
    End If
    ! Open the Energy.dat file
    Open(unit=24,file="Energy.dat",status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_AlphaP!"  
       Write(*,*)"Could not open Energy.dat file."
       Stop
    End If

    ! Open the Output file
    Open(27,File=trim(seedname)//"_alphap.dat",status='replace',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_AlphaP!"
       Write(*,*)"Could not open "//trim(seedname)//"_alphap.dat for output"
       Stop
    End If

    Write(27,*)"# Thermal expansion coefficient for CASTEP run "//trim(Seedname)
    Write(27,*)"# Calculated using mdtep."

    ! Initial block size is one to obtain the total
    ! averages - used to calculate central Cp 
    ! value
    CorrNSamp = 0

    ! We need the average pressure to continue
    MeanP = 0.0_dp

    Do I = 1,Start -1
       Read(23,*,end=1,err=1)Dummy,Dummy,Pressure,Dummy
       MeanP = MeanP + Pressure
    End Do

    Do I = Start,Final
       Read(23,*,end=1,err=1)Dummy,Dummy,Pressure,Dummy
       MeanP = MeanP + Pressure
    End Do

1   Rewind(23)

    MeanP = MeanP/Real(N,kind=dp)
    MeanP = MeanP/29414000.0_dp

    EnthalpMean = 0.0_dp
    VolMean     = 0.0_dp

    ! We now need the average enthalpy and volume
    ! in order to average over deviations from these.
    Do I = 1,Start - 1
       Read(23,*)dummy,dummy,dummy,Volume
       Read(24,*)dummy,dummy,dummy,TotalE
    End Do
    Do I = Start,Final
       Read(23,*)dummy,dummy,dummy,Volume
       Read(24,*)dummy,dummy,dummy,TotalE
       EnthalpMean = EnthalpMean + TotalE + MeanP*Volume
       Volmean     = Volmean + Volume
    End Do

    EnthalpMean = EnthalpMean/Real(N,kind=dp)
    Volmean     = Volmean/Real(N,kind=dp)

    Rewind(23)
    Rewind(24)

    Do           ! Loop over block sizes 

       CorrNSamp = 1+CorrNSamp    ! increment block size 

       ! Rewind the input file from last block size
       Rewind(23)
       Rewind(24)

       ! Counter for reaching blocksize
       CorrCount = 0

       ! Zero all accumulators
       ProdMean     = 0.0_dp
       TMean        = 0.0_dp
       Tmeansq      = 0.0_dp
       Prodmeansq   = 0.0_dp
       Volmean2     = 0.0_dp
       Volmeansq    = 0.0_dp

       ! Zero accumulators within block
       BlockTemp   = 0.0_dp
       BlockVol    = 0.0_dp
       BlockProd   = 0.0_dp

       N = 0

       Do I = 1, Start-1
          Read(23,*,end=5,err=5)dummy,temp,dummy,Volume
          Read(24,*)dummy,dummy,dummy,TotalE
       End Do
          
       ! Loop over all entries in Sample.dat
       Do I = Start,Final
          ! Read in the temperature and volume
          Read(23,*)dummy,temp,dummy,Volume
          Read(24,*)dummy,dummy,dummy,TotalE

          ! Calculate Enthalpy
          Enthalpy = TotalE + MeanP*Volume

          ! Average within the current block
          CorrCount   = CorrCount + 1

          BlockTemp   = BlockTemp + Temp

          BlockProd   = BlockProd + (Enthalpy-EnthalpMean) &
               *(Volume-Volmean)

          BlockVol    = BlockVol  + Volume

          ! If we have reached the end of a block then
          ! accumumate the global averages
          If (CorrCount == CorrNSamp) then

             ! Average temp and volume for the preceeding block
             Temp   = BlockTemp/Real(CorrNSamp,kind=dp)
             Prod   = BlockProd/Real(CorrNSamp,kind=dp)
             Volume = BlockVol /Real(CorrNSamp,kind=dp)

             ! Zero the within-block averages
             BlockTemp = 0.0_dp
             BlockProd = 0.0_dp
             BlockVol  = 0.0_dp

             ! Accumunate the global means
             Tmean = Tmean + Temp
             ProdMean = ProdMean + Prod
             VolMean2 = Volmean2 + Volume

             ! And the global mean-squares
             Tmeansq = Tmeansq + Temp*Temp
             ProdMeansq = ProdMeansq + Prod*Prod
             VolMeansq  = Volmeansq + Volume*Volume

             ! Count the number of blocks so far
             N = N +1

             ! Reset the block position counter
             CorrCount =0

          End If

       End Do   ! End loop over entries in sample file

5      Continue 

       ! If the number of blocked samples is less than
       ! 100 then get out.
       If (N.LT.10) then
          Write(27,*)
          Write(27,*)
          Write(27,*)"Less than 10 blocks - Stopping"
          Close(27)
          Close(23)
          Close(24)
          return
       End If

       ! Calculate the mean and the error
       ! in the mean for both the temperature
       ! and volume in the usual way.

       Tmean = Tmean / Real(N,kind=dp)
       Tmeansq = Tmeansq / Real(N,kind=dp)

       Tsd  = SQRT(Tmeansq-Tmean*Tmean)
       Terr = Tsd/Real(N,kind=dp)

       ProdMean = Prodmean / Real(N,kind=dp)
       ProdMeansq = ProdMeansq / Real(N,kind=dp)

       Prodsd = SQRT(ProdMeansq-Prodmean*Prodmean)
       Proderr = Prodsd/Real(N,kind=dp)

       Volmean2 = Volmean2/Real(N,kind=dp)
       Volmeansq = volmeansq/Real(N,kind=dp)

       Volsd  = SQRT(Volmeansq-volmean2*volmean2)
       Volerr = Volsd/Real(N,kind=dp) 

       ! Calculate alpha_p -- only printed for block size of one.

       Alpha_p = prodmean/(Kb*Tmean*Tmean*Volmean2)

       If (FirstPass) then 
          Write(27,*)
          Write(27,*)"Alpha_{P} = ",Alpha_p,"per Kelvin"
          Write(27,*)
          Write(27,'(2A20)')"Block Size","Error in C_{P}"
          Write(27,*)"========================================"
          FirstPass = .FALSE.
       End If

       ! Calculate error in alpha_p for the current blocksize
       alpha_perr = proderr/(Kb*Tmean*Tmean*Volmean) & 
            -2.0_dp * Prod*Terr/(Tmean**3*Volmean) &
            - prod*Volerr/(Kb*(Tmean*Volmean**2))


       Write(27,'(I20,E20.8)')CorrNSamp,alpha_perr

    End Do ! End Loop over block sizes.

    Close(27)
    Close(23)
    Close(24)

    Return

  End Subroutine Analyse_AlphaP

  Subroutine Analyse_TempTest(Start,Final)
    !=========================================================================!
    ! Calculates the distribution of temperatures and compares to the         !
    ! canonical distribution for a system coupled to a heat bath!             !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !	Start    , Intent IN  , Timestep at which averaging is to begin       !
    !   Final    , Intent IN  , Timestep at which averaging is to end         !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !                Num_Atoms   , Number of atoms in the output              !
    !                seedname    , name for the castep run under analysis     !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! A  Sample.dat file should exist in the current directory                !
    !                                                                         !
    ! NOTE                                                                    !
    !                                                                         !
    ! To correctly calculate the canonical temperature distribution we must   !
    ! know the number of ionic constraints. This information is not in the    !
    ! md file. We therefore first try to open seedname.castep to find the     !
    ! constraint information. If this file doesn't exist or doesn't contain   !
    ! the information then we assume there are no ionic constraints.          !
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================!  
    Use Global   , Only : MD_Steps, Num_Atoms, seedname
    Use Constants, Only : kb
    Implicit None

    Integer(kind=int32),Intent(IN) :: Start
    Integer(kind=int32),Intent(IN) :: Final

    Integer(kind=int32) :: I,N,Ndeg,J,ierr,Ddeg,Nc

    Real(kind=dp)       :: Startbin,EndBin,BinWidth,UpperLim
    Real(kind=dp)       :: Av,SDDev,Analytic,Dummy,Temp

    Integer(kind=int32),Parameter   :: NumBins = 50

    Integer(kind=int32),Dimension(NumBins) :: Bin

    ! Dummy string
    Character(25) :: DString25
    Character(29) :: DString29

    ! Zero the bin array
    Bin = 0

    ! Checks
    If (Final>MD_Steps) then
       Write(*,*)"Error in Analyse_TempTest!"
       Write(*,*)"Final sample point is beyond end of run."
       Stop
    End If
    If (Start<0.OR.Start>MD_Steps) then
       Write(*,*)"Error in Analyse_TempTest!"
       Write(*,*)"Start point is invalid."
       Stop
    End If
    If (Start>Final) then
       Write(*,*)"Error in Analyse_TempTest!"
       Write(*,*)"End pos is before start pos."
       Stop
    End If


    ! Set Ndeg initially equal to 3xNum_Atoms
    Ndeg = 3*Num_Atoms

    Nc = -1

    ! Can we open a castep file?
    Open(unit=24,file=trim(seedname)//".castep",status='old',iostat=ierr)
    If (ierr/=0) then
       ! We can only assume no ionic constraints
       Write(*,*)
       Write(*,*)"WARNING - no file named "//trim(seedname)//".castep was found."
       Write(*,*)"Assuming no ionic constraints."
       Write(*,*)
       Nc = 0
    Else
       ! Loop over entries in castep file
       Do
          ! Try and read a constraints line
          Read(24,'(a25,a29,i8)',iostat=ierr,end=1)DString25,DString29,Nc
          If (ierr==0) then
             ! Double check that this is a constraints lline
             If (Dstring29=="Number of ionic constraints =") then
                Write(*,*)
                Write(*,*)"Identified ",Nc," ionic constraints in "//trim(seedname)//".castep"
                Write(*,*)
                Ndeg = Ndeg - Nc   ! Adjust Ndeg
                Exit
             End If
          End If
       End Do

1      Continue

       If (Nc==-1) then
          Write(*,*)
          Write(*,*)"WARNING - no constraint information was found in "//trim(seedname)//".castep"
          Write(*,*)
       End If

       Close(24)

    End If

    !Open Sample .dat from the MD program. 
    Open(23,File='Sample.dat',status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_TempTest!"  
       Write(*,*)"Could not open Sample.dat file."
       Stop
    End If

    !Initialise StartBin and EndBin
    StartBin = 9999999.9_dp
    EndBin = -99999999.9_dp
    N = 0

    ! Loop over unwanted entries
    Do I = 1,Start-1
       Read(23,*)Dummy,Temp,Dummy,Dummy
    End Do

    !Count records and find max and min
    Do I = Start,Final
       Read(23,*)Dummy,Temp,Dummy,Dummy
       If (Temp.LT.StartBin) Startbin = Temp
       If (Temp.GT.EndBin)   Endbin   = Temp
    End Do

    N = Final-Start


    !Want all temperatures to lie WITHIN that range.
    StartBin = 0.9_dp*StartBin
    EndBin   = 1.1_dp*EndBin

    !Set the width of each bin
    BinWidth = (EndBin-StartBin)/REAL(NumBins,kind=dp)

    !Go back to the start of the file
    Rewind(23)

    !Initialise the average temperature
    av = 0.0_dp

    ! Loop over unwanted entries
    Do I = 1,Start-1
       Read(23,*,Err=1,End=1)Dummy,Temp,Dummy,Dummy
    End Do

    Do J = Start,Final

       Read(23,*)Dummy,Temp,Dummy,Dummy

       !Increment the bin this temperature falls into
       I = Int((Temp-StartBin)*REAL(NumBins,kind=dp)/(EndBin-StartBin)) + 1
       Bin(I) = Bin(I) + 1

       !Acumulate the average temperature
       av = av + temp

    End Do

    !Calculate the average temperature and the 
    !ideal standard deviation.
    av = av/REAL(N,kind=dp)
    SDDev = av*SQRT(2.0_dp/(REAL(NDeg,kind=dp)))

    !Open the output file
    Open(27,File=trim(seedname)//'_TDIST.dat',status='replace',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_temptest!"
       Write(*,*)"Could not open "//trim(seedname)//"_TDIST.dat for output"
       Stop
    End If

    Write(27,*)"# Temperature distribution for CASTEP run "//trim(Seedname)
    Write(27,*)"# Calculated using mdtep."


    Do I =1,NumBins
       !Calculate the mid-bin temperature
       Temp = Startbin+(Real((I-0.5),kind=dp)*BinWidth)

       !Calculate the normalised ideal prob for this T
       Analytic = 1.0_dp/(SDDev*SQRT(2.0_dp*3.14159_dp))
       Analytic = Analytic*EXP(-(Temp-av)**2/(2.0_dp*sddev**2))

       !Write the normalised bin count & ideal value to file
       Write(27,*)Temp,Bin(I)/(BinWidth*N),Analytic
    End Do

    Close(27)
    Close(23)

    Return

  End Subroutine Analyse_TempTest

  Subroutine Analyse_VolTest(Start,Final)
    !=========================================================================!
    ! Calculates volume distribution and compares to exact canonical.         !
    !-------------------------------------------------------------------------!
    ! Arguments:                                                              !
    !	Start    , Intent IN  , Timestep at which averaging is to begin       !
    !   Final    , Intent IN  , Timestep at which averaging is to end         !
    !-------------------------------------------------------------------------!
    ! Parent module variables used:                                           !
    ! LocalBT    , Stored value of BetaT from earler call to BetaT routine    !
    !-------------------------------------------------------------------------!
    ! Modules used:                                                           !
    ! Global      ,  MD_Steps    , Total Number of MD Steps in the output     !
    !                Num_Atoms   , Number of atoms in the output              !
    !                seedname    , name for the castep run under analysis     !
    !-------------------------------------------------------------------------!
    ! Key Internal Variables:                                                 !
    !-------------------------------------------------------------------------!
    ! Necessary conditions:                                                   !
    ! A  Sample.dat file should exist in the current directory                !
    !                                                                         !
    ! NOTE                                                                    !
    !                                                                         !
    ! As we do not have knowledge of the compressibility we must have         !
    ! previously called Analyse_BetaT. Hence the plotted 'ideal' solution     !
    ! is actually a Gaussian fit to the simulation data.                      !              
    !-------------------------------------------------------------------------!
    ! Written by D. Quigley Jan 2004                                          !
    !=========================================================================! 
    Use Constants
    Use Global   , Only : MD_Steps, Num_Atoms, seedname
    Use Constants, Only : kb
    Implicit None

    Integer(kind=int32),Intent(IN) :: Start,Final

    Integer(kind=int32),Parameter   :: NumBins = 50

    !Array to hold the HitCounter
    Integer(kind=int32),Dimension(NumBins) :: Bin
 
    Integer(kind=int32) :: I,N,Natoms,J,ierr
    
    Real(kind=dp)       :: Startbin,EndBin,BinWidth,UpperLim
    Real(kind=dp)       :: Av,SDDev,Analytic,Dummy,Volume
    Real(kind=dp)       :: BetaT,Temp,Avtemp


    ! Checks
    If (Final>MD_Steps) then
       Write(*,*)"Error in Analyse_Voltest!"
       Write(*,*)"Final sample point is beyond end of run."
       Stop
    End If
    If (Start<0.OR.Start>MD_Steps) then
       Write(*,*)"Error in Analyse_Voltest!"
       Write(*,*)"Start point is invalid."
       Stop
    End If
    If (Start>Final) then
       Write(*,*)"Error in Analyse_Voltest!"
       Write(*,*)"End pos is before start pos."
       Stop
    End If


    ! Check that the module level compressibility is set
    If (Analyse_LocalBT<0.0_dp) then
       Write(*,*)
       Write(*,*)"Calculating Beta_T to fit Gaussian to volume samples."
       Write(*,*)
       Call Analyse_BetaT(Start,Final)
    End If

    BetaT = Analyse_LocalBT

    BetaT = 1.0d0/(1000.0d0*BetaT)

    BetaT = BetaT*29414000.0

    ! Open Sample .dat from the MD program. 
    Open(23,File='Sample.dat',status='old',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_TempTest!"  
       Write(*,*)"Could not open Sample.dat file."
       Stop
    End If

    !Initialise StartBin and EndBin
    StartBin =  99999999.9_dp
    EndBin   = -99999999.9_dp
    N = 0

    ! Skip unwanted data
    Do I = 1,Start-1
       Read(23,'(4F15.6)')Dummy,Dummy,Dummy,Volume
    End Do

    ! Find max and min
    Do I = Start,Final
       Read(23,'(4F15.6)')Dummy,Dummy,Dummy,Volume
       N = N+1
        If (Volume.LT.StartBin) Startbin = Volume
       If (volume.GT.EndBin)   Endbin   = Volume
    End Do


    !Want all temperatures to lie WITHIN that range.
    StartBin = StartBin - (0.5_dp/Real(NumBins,kind=dp))*(EndBin-StartBin)
    EndBin   = EndBin + (0.5_dp/Real(NumBins,kind=dp))*(EndBin-StartBin)
    !Set the width of each bin
    BinWidth = (EndBin-StartBin)/REAL(NumBins)
    
    !Go back to the start of the file
    Rewind(23)

    !Initialise the average temperature
    av = 0.0_dp
    avtemp = 0.0_dp

    ! Skip unwanted data
    Do I = 1,Start-1
       Read(23,*)Dummy,Dummy,Dummy,Volume
    End Do

    Do J = Start,Final

        Read(23,*)Dummy,Temp,Dummy,Volume

        !Increment the bin this temperature falls into
        I = Int((Volume-StartBin)*REAL(NumBins,kind=dp)/(EndBin-StartBin)) + 1
        Bin(I) = Bin(I) + 1

        !Acumulate the average temperature
        av = av + Volume
        avTemp = avTemp + Temp

    End Do
    
    !Calculate the average temperature and the 
    !idea standard deviation.
    av = av/REAL(N,kind=dp)
    avTemp = AvTemp/Real(N,kind=dp)


    SDDev = SQRT(av*3.166874348e-6_dp*AvTemp*BetaT)

  
    !Open the output file
    Open(27,File=trim(seedname)//'_VDIST.dat',status='replace',iostat=ierr)
    If (ierr/=0) then
       Write(*,*)"Error in Analyse_temptest!"
       Write(*,*)"Could not open "//trim(seedname)//"_VDIST.dat for output"
       Stop
    End If

    Write(27,*)"# Volume distribution for CASTEP run "//trim(Seedname)
    Write(27,*)"# Calculated using mdtep."

    Do I =1,NumBins
        !Calculate the mid-bin temperature
        Volume = Startbin+((I-0.5)*BinWidth)
        
        !Calculate the normalised ideal prob for this T
        Analytic = 1.0_dp/(SDDev*SQRT(2.0_dp*3.14159_dp))
        Analytic = Analytic*EXP(-(Volume-av)**2/(2.0_dp*sddev**2))
        
        !Write the normalised bin count & ideal value to file
        Write(27,*)Volume,Real(Bin(I),kind=dp)/Real((BinWidth*N),kind=dp),Analytic
    End Do

    Close(27)
    Close(23)

    Return

  End Subroutine Analyse_Voltest


End Module Analyse
